# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import time
import multiprocessing
from pyqode.core.widgets import (
    FileSystemTreeView,
    FileSystemContextMenu
)
from qtpy.QtWidgets import (
    QCheckBox,
    QDockWidget,
    QWidget,
    QVBoxLayout,
    QApplication
)
from qtpy.QtCore import QFileSystemWatcher, QTimer
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'OpenSesameIDE', category=u'extension')


class FolderBrowserDockWidget(QDockWidget):

    def __init__(self, parent, ide, path):

        super(FolderBrowserDockWidget, self).__init__(parent)
        self.main_window = parent
        self._ide = ide
        self.path = path
        self._gitignore_checkbox = QCheckBox(_('Use .gitignore'))
        self._gitignore_checkbox.setChecked(True)
        self._folder_browser = FolderBrowser(parent, ide, path, self)
        self._container_layout = QVBoxLayout()
        self._container_widget = QWidget(self)
        self._container_layout.setContentsMargins(6, 6, 6, 6)
        self._container_layout.addWidget(self._folder_browser)
        self._container_layout.addWidget(self._gitignore_checkbox)
        self._container_widget.setLayout(self._container_layout)
        self.setWidget(self._container_widget)
        self.setWindowTitle(os.path.basename(path))
        self.setObjectName(os.path.basename(path))

    def select_path(self, path):

        self._folder_browser.select_path(path)

    def closeEvent(self, e):

        self._folder_browser.shutdown()
        self._ide.remove_folder_browser_dock_widget(self)
        self.close()

    @property
    def file_list(self):

        return self._folder_browser.file_list


class FolderBrowser(FileSystemTreeView):

    def __init__(self, parent, ide, path, dock_widget):

        super(FolderBrowser, self).__init__(parent)
        self.main_window = parent
        self._path = path
        self._ide = ide
        self._dock_widget = dock_widget
        self._dock_widget._gitignore_checkbox.stateChanged.connect(
            self._toggle_gitignore
        )
        self._set_ignore_patterns()
        self.set_root_path(os.path.normpath(path))
        self.set_context_menu(FileSystemContextMenu(self.main_window))
        self._watcher = QFileSystemWatcher()
        self._watcher.addPath(path)
        self._watcher.fileChanged.connect(self._on_file_changed)
        self._watcher.directoryChanged.connect(self._on_folder_changed)
        self._indexing = False
        self._file_list = []
        self._active = True
        self._timer = None
        self._index_files()
        
    def _refresh(self):
        """Refreshes the file tree"""
        self.set_root_path(self.root_path)
        
    def _toggle_gitignore(self):
        """Is called when the .gitignore option is (un)checked. Updates the
        ignore patterns and refreshes the file tree.
        """
        self._set_ignore_patterns()
        self._restart_indexing()
        self._refresh()
        
    def _set_ignore_patterns(self):
        """Sets the ignore patterns based on the configuration as well as the
        .gitignore if available and activated.
        """
        self.clear_ignore_patterns()
        self.add_ignore_patterns(self._ide.ignore_patterns)
        gitignore = os.path.join(self._path, u'.gitignore')
        if not os.path.exists(gitignore):
            self._dock_widget._gitignore_checkbox.hide()
            return
        self._dock_widget._gitignore_checkbox.show()
        if self._dock_widget._gitignore_checkbox.isChecked():
            with open(gitignore) as fd:
                self.add_ignore_patterns(fd.read().splitlines())

    def currentChanged(self, current_index, previous_index):
        """If the current item has changed, then the container dock widget
        should be raised. This is necessary for locating files in dock widgets
        that are tabbed and not active.
        """
        super(FolderBrowser, self).currentChanged(
            current_index,
            previous_index
        )
        self._dock_widget.raise_()

    def mouseDoubleClickEvent(self, event):
        """Opens a file when it's double-clicked"""
        path = self.fileInfo(self.indexAt(event.pos())).filePath()
        self.open_file(path)
        
    def shutdown(self):
        
        self._active = False
        self._watcher.fileChanged.disconnect()
        self._watcher.directoryChanged.disconnect()
        self._stop_indexing()

    def open_file(self, path):

        if not os.path.exists(path) or os.path.isdir(path):
            return
        self._ide.open_document(path)

    @property
    def file_list(self):

        while self._indexing:
            oslogger.debug(u'still indexing {}'.format(self._path))
            time.sleep(.1)
            QApplication.processEvents()
        return self._file_list
    
    def _on_file_changed(self, _=None):
        
        oslogger.debug(u'file changed in {}'.format(self._path))
        self._restart_indexing()

    def _on_folder_changed(self, _=None):
        
        oslogger.debug(u'folder changed in {}'.format(self._path))
        self._restart_indexing()
        
    def _stop_indexing(self):
        """Stops the file indexer by stopping any active timers and by
        killing running indexer processes.
        """
        if self._timer is not None and self._timer.isActive():
            oslogger.debug(u'stopping timer for {}'.format(self._path))
            self._timer.stop()
        if self._indexing:
            oslogger.debug(u'killing indexer for {}'.format(self._path))
            self._file_indexer.terminate()
            self._indexing = False
        
    def _restart_indexing(self):
        """Restarts the file indexer"""
        self._stop_indexing()
        self._index_files()
        
    def _start_timer(self, msec, target):
        self._timer = QTimer()
        self._timer.setSingleShot(True)
        self._timer.setInterval(msec)
        self._timer.timeout.connect(target)
        self._timer.start()
        
    def _index_files(self, _=None):

        if not self._active:
            oslogger.debug(u'shutting down indexing for {}'.format(self._path))
            return
        if self._indexing:
            oslogger.debug(u'indexing in progress for {}'.format(self._path))
            return
        self._indexing = True
        self._queue = multiprocessing.Queue()
        self._file_indexer = multiprocessing.Process(
            target=file_indexer,
            args=(
                self._queue,
                self._path,
                self._ignored_patterns,
                cfg.opensesame_ide_max_index_files
            )
        )
        self._file_indexer.start()
        oslogger.debug(u'indexing {} (PID={})'.format(
            self._path,
            self._file_indexer.pid
        ))
        self._ide.extension_manager.fire(
            'register_subprocess',
            pid=self._file_indexer.pid,
            description='file_indexer:{}'.format(self._path)
        )
        self._start_timer(1000, self._check_file_indexer)

    def _check_file_indexer(self):

        if self._queue.empty():
            oslogger.debug(u'queue still empty for {}'.format(self._path))
            self._start_timer(1000, self._check_file_indexer)
            return
        self._file_list = self._queue.get()
        self._indexing = False
        if self._file_list is None:
            self._ide.extension_manager.fire(
                u'notify',
                message=_(u'Not indexing {} (too many files)').format(
                    self._path
                ),
                category=u'info'
            )
            self._file_list = []
            return
        oslogger.debug(u'{} files indexed for {}'.format(
            len(self._file_list),
            self._path)
        )
        self._file_indexer.join()
        try:
            self._file_indexer.close()
        except AttributeError:
            # Process.close() was introduced only in Python 3.7
            pass
        self._start_timer(300000, self._index_files)


def file_indexer(queue, path, ignore_patterns, max_files):
    
    from pathspec import PathSpec

    def _list_files(dirname):

        files = []
        try:
            basenames = os.listdir(dirname)
        except Exception:
            pass
        else:
            for basename in basenames:
                path = os.path.join(dirname, basename)
                if ignore_spec.match_file(path):
                    continue
                if os.path.isdir(path):
                    files += _list_files(path)
                else:
                    files.append(path)
        if len(files) > max_files:
            raise ValueError(u'Too many files')
        return files

    ignore_spec = PathSpec.from_lines('gitwildmatch', ignore_patterns)
    try:
        files = _list_files(path)
    except ValueError:
        files = None
    queue.put(files)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import math
from PIL.ImageQt import ImageQt
from qtpy.QtGui import QPixmap
from qtpy.QtCore import QSize
from qtpy.QtWidgets import QLabel


class ImageLabel(QLabel):
    """Implements a label with a resizable image that maintains its original
    aspect ratio.
    """
    
    def __init__(self, img):
        
        QLabel.__init__(self)
        self._ratio = img.size[0] / img.size[1]
        self._pixmap = QPixmap.fromImage(ImageQt(img))
        self.setMinimumSize(QSize(10, 10))
        self.setScaledContents(True)
        self.setPixmap(self._pixmap)
        
    def resizeEvent(self, e):

        size = e.size()
        height = size.height()
        width = size.width()
        ratio = width / height
        if math.isclose(ratio, self._ratio):
            return
        if ratio > self._ratio:
            new_size = QSize(height * self._ratio, height)
        elif ratio < self._ratio:
            new_size = QSize(width, width / self._ratio)
        self.resize(new_size)


def inspect_PngImageFile(value):

    return ImageLabel(value)



#!/usr/bin/env python3
# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys
import fnmatch
import json
from setuptools import setup


EXCLUDE = [
    u'*~',
    u'*.pyc',
    u'*.pyo',
    u'*__pycache__*'
]

EXTENSIONS = [
    u'IDELauncher',
    u'OpenSesameIDE',
    u'FindInFiles',
    u'JupyterConsole',
    u'JupyterNotebook',
    u'QuickSelector',
    u'WorkspaceExplorer',
    u'SymbolSelector',
    u'RapunzelWelcome',
    u'CommandPalette',
    u'WordCount',
    u'SpellCheck',
    u'DataViewer',
    u'GitGUI',
    u'RapunzelLocale',
    u'PythonDebugger',
    u'SubprocessManager',
    u'ImageAnnotations'
]


def is_excluded(path):

    return any(fnmatch.fnmatch(path, m) for m in EXCLUDE)


def recursive_glob(src_folder, target_folder):

    """
    desc:
        Recursively gets all files that are in src folder.

    arguments:
        src_folder:       The source folder.
        target_folder:    The target folder.

    returns:
        A list of (target folder, filenames) tuples.
    """

    globbed = []
    path_list = []
    for path in os.listdir(src_folder):
        full_path = os.path.join(src_folder, path)
        if is_excluded(full_path):
            continue
        if os.path.isdir(full_path):
            globbed += recursive_glob(
                full_path,
                os.path.join(target_folder, path)
            )
            continue
        path_list.append(full_path)
    globbed.append((os.path.join(u'share', target_folder), path_list))
    return globbed


def extensions():

    """
    desc:
        Create a list of all extension files that should be included.

    returns:
        A list of (target folder, filenames) tuples.
    """

    globbed = []
    for extension in EXTENSIONS:
        folder = os.path.join(u'opensesame_extensions', extension)
        globbed += recursive_glob(folder, folder)
    return globbed


def data_files():

    return (
        [
            (u"share/icons/hicolor/scalable/apps", [u"mime/rapunzel.svg"]),
            (u"share/applications", [u"mime/rapunzel.desktop"]),
        ] +
        extensions()
    )


def get_version():

    with open('opensesame_extensions/OpenSesameIDE/info.json') as fd:
        info = json.load(fd)
    return info['version']


def get_readme():

    if os.path.exists('readme.md'):
        with open('readme.md') as fd:
            return fd.read()
    return 'No readme information'


setup(
    name='rapunzel',
    version=get_version(),
    description='Turns OpenSesame into a Python code editor',
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    author='Sebastiaan Mathot',
    author_email='s.mathot@cogsci.nl',
    url='https://github.com/smathot/rapunzel',
    entry_points={
        'gui_scripts': [
            'rapunzel = rapunzel:rapunzel'
        ]
    },
    classifiers=[
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    py_modules=['rapunzel'],
    # The dependencies are ignored when invoked by stdeb3
    install_requires=(
        [] if 'install' in sys.argv and '--root' in sys.argv
        else [
            'python-levenshtein',
            'python-opensesame',
            'qtconsole',
            'nbformat',
            'pyspellchecker',
            'esprima',
            'psutil'
        ]
    ),
    data_files=data_files()
)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QDialog,
    QLineEdit,
    QVBoxLayout,
    QListWidget,
    QListWidgetItem
)
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
from libqtopensesame.misc.config import cfg
_ = translation_context(u'QuickSelector', category=u'extension')

STYLESHEET = '''
QuickSelectorDialog {{
    background-color: {background};
    color: {foreground};
    padding: 8px;
    border-radius: 4px;
}}
QLineEdit {{
    margin-bottom: 4px;
}}
QLineEdit, QListWidget {{
    color: {foreground};
    border: none;
    font-family: {font_family};
    font-size: {font_size}px;
    background-color: {selected_background};
    padding: 4px;
}}
'''


class SearchBox(QLineEdit):

    def keyPressEvent(self, e):

        if e.key() in (Qt.Key_Down, Qt.Key_Home):
            self.parent().focus_result_box()
            return
        if e.key() in (Qt.Key_Up, Qt.Key_End):
            self.parent().focus_result_box(jump_to_last=True)
            return
        if e.key() in (Qt.Key_Enter, Qt.Key_Return):
            self.parent().select_top_result()
            return
        super(SearchBox, self).keyPressEvent(e)


class ResultBox(QListWidget):

    def keyPressEvent(self, e):

        if (
            (e.key() == Qt.Key_Up and not self.currentRow()) or
            e.key() not in (
                Qt.Key_Down,
                Qt.Key_Up,
                Qt.Key_Return,
                Qt.Key_Enter,
                Qt.Key_PageUp,
                Qt.Key_PageDown
            )
        ):
            self.parent().focus_search_box(e)
            return
        super(ResultBox, self).keyPressEvent(e)


class QuickSelectorDialog(QDialog):

    def __init__(self, parent, haystack, placeholder_text, default):

        from pygments import styles, token

        super(QuickSelectorDialog, self).__init__(
            parent,
            Qt.FramelessWindowHint
        )
        self.setWindowTitle(u'Quick Selector')
        style = styles.get_style_by_name(cfg.pyqode_color_scheme)
        self.setStyleSheet(STYLESHEET.format(
            background=style.highlight_color,
            selected_background=style.background_color,
            foreground=style.styles.get(token.Text),
            font_family=cfg.pyqode_font_name,
            font_size=cfg.pyqode_font_size
        ))
        # Already create a lowercase version of the label for performance
        self._haystack = [
            (label.lower(), label, data, on_select)
            for label, data, on_select in haystack
        ]
        self._default = default
        self._search_box = SearchBox(self)
        self._search_box.textEdited.connect(self._search)
        if placeholder_text:
            self._search_box.setPlaceholderText(placeholder_text)
        self._result_box = ResultBox(self)
        self._result_box.setTextElideMode(Qt.ElideMiddle)
        self._result_box.itemActivated.connect(self._select)
        self._result_box.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self._result_box.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self._layout = QVBoxLayout(self)
        self._layout.addWidget(self._search_box)
        self._layout.addWidget(self._result_box)
        self.setMinimumWidth(cfg.quick_selector_min_width)
        self._search(u'')

    def focus_result_box(self, jump_to_last=False):

        self._result_box.setFocus()
        self._result_box.setCurrentRow(
            self._result_box.count() - 1 if jump_to_last else 0
        )

    def focus_search_box(self, e):

        self._search_box.setFocus()
        self._search_box.keyPressEvent(e)

    def select_top_result(self):

        self._select(self._result_box.takeItem(0))

    def _search(self, needle):

        import Levenshtein

        if not needle:
            haystack = self._haystack[:]
            if self._default:
                haystack.insert(0, (self._default[0].lower(),) + self._default)
            self._result_box.clear()
            for lower_label, label, data, on_select in haystack:
                item = QListWidgetItem(label, self._result_box)
                item.data = data
                item.on_select = on_select
                self._result_box.addItem(item)
            return
        needle = needle.lower()
        needles = needle.split()
        self._result_box.show()
        self._result_box.clear()
        results = []
        for lower_label, label, data, on_select in self._haystack:
            for subneedle in needles:
                if subneedle not in lower_label:
                    break
            else:
                # We count the number of full matches of the needle in the
                # labels and divide the score by this. This was we give an
                # advantage to labels that contain multiple matches of the
                # needle, for example in the folder and file name.
                hits = max(
                    1,
                    sum(lower_label.count(s) for s in needle.split())
                )
                results.append((
                    Levenshtein.distance(needle, lower_label) / (len(label) * hits),
                    label,
                    data,
                    on_select
                ))
        for score, label, data, on_select in sorted(
            results,
            key=lambda t: t[0]
        ):
            item = QListWidgetItem(label, self._result_box)
            item.data = data
            item.on_select = on_select
            self._result_box.addItem(item)

    def _select(self, item):

        if item is None:
            return
        item.on_select(item.data)
        self.accept()


class QuickSelector(BaseExtension):

    def event_quick_select(
        self, haystack,
        placeholder_text=None,
        default=None
    ):

        QuickSelectorDialog(
            self.main_window,
            haystack,
            placeholder_text,
            default
        ).exec_()
